//
//  iso9141_io.c
//  HA7 logger mini
//  Copyright(C)2012 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//    Update history
//    ---------- ----- -------------------------------------------
//    2012.01.02 v0.0  First cording
//    2012.01.12 v0.1  Commit
//    2012.02.29 v0.1a Bug fix
//

#include "iso9141_io.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include "ftd2xx.h"

#define P1_MIN		0			// [us]
#define P2_MIN		25000		// [us]
#define P2_MAX		50000		// [us]
#define P3_MIN		55000		// [us]
#define P4_MIN		5000		// [us]

#define SYNC_TRY	100			// max 300ms (300ms/50ms)

FT_HANDLE	fd;

int iso9141_open(char *dev){
	FT_STATUS	rc;

	rc = FT_OK;
	while(1) {
		if((rc = FT_ListDevices((PVOID)0, dev, FT_LIST_BY_INDEX | FT_OPEN_BY_SERIAL_NUMBER)) != FT_OK) {
			break;
		}
	
		if (fd == NULL) {
			if((rc = FT_Open(0, &fd)) != FT_OK){
				break;
			}
		}
	
//		if ((rc = FT_SetBaudRate(fd, 10500)) != FT_OK) {	// DEL 2012.2.29
		if ((rc = FT_SetBaudRate(fd, 10400)) != FT_OK) {	// ADD 2012.2.29
			break;
		}
	
		if ((rc = FT_SetDataCharacteristics(fd, FT_BITS_8, FT_STOP_BITS_1, FT_PARITY_NONE)) != FT_OK) {
			break;
		}
	
		if ((rc = FT_SetFlowControl(fd, FT_FLOW_NONE, 0, 0)) != FT_OK) {
			break;
		}
	
		if ((rc = FT_SetTimeouts(fd, P2_MAX/1000, P2_MAX/1000)) != FT_OK) {
			break;
		}
		break;
	}
	if (rc != FT_OK) {
		FT_Close(fd);
		fd = NULL;
	}
	return((int)rc);
}

int iso9141_close(void){
	FT_STATUS	rc;
	
	rc = FT_OK;
	if (fd != NULL) {
		if ((rc = FT_Close(fd)) == FT_OK) {
			fd = NULL;
		}
	}
	return((int)rc);
}

FT_STATUS send_5baud(unsigned char *dat, size_t len){
	FT_STATUS	rc;
	useconds_t	wt = (useconds_t)(1.0/5*1000000);	// 5bps
	
	for(; len > 0; --len) {
		char	c;
		int	i;
		
		c = *dat++;
		
		// Start bit
		if ((rc = FT_SetBreakOn(fd)) != FT_OK) {
			break;
		}
		usleep(wt);
		
		// Data bits
		for(i = 0; i < 8; i++) {
			if (c & 0x01) {
				rc = FT_SetBreakOff(fd);
			} else {
				rc = FT_SetBreakOn(fd);
			}
			if (rc != FT_OK) {
				break;
			}
			c = c >> 1;
			usleep(wt);
		}
		if (rc != FT_OK) {
			break;
		}
		
		// Stop bit
		if ((rc = FT_SetBreakOff(fd)) != FT_OK) {
			break;
		}
		usleep(wt);
	}
	
	return(rc);
}

int iso9141_init(unsigned char keyword[2]){
	FT_STATUS	rc;
	unsigned char	addr	= 0x33;
	
	do {
		DWORD	l;
		int	i;
		unsigned char	c;
		
		// 5 baud method
		usleep(2000);
		rc = send_5baud(&addr, sizeof(addr));
		
		// Recieve sync
		for(i = 0; i < SYNC_TRY; i++) {
			if ((rc = FT_Read(fd, &c, 1, &l)) != FT_OK) {
				break;
			}
			if ((l > 0) && (c == 0x55)) {
				break;
			}
		}
		if (rc != FT_OK) {
//			printf("iso9141_init: sync recv err rc=%d\n", rc);
			break;
		}
		if (i >= SYNC_TRY) {
			rc = FT_OTHER_ERROR;		// Time out
//			printf("iso9141_init: sync recv time out rc=%d\n", rc);
			break;
		}
		
		// Recieve key word bytes
		if ((rc = FT_Read(fd, keyword, sizeof(keyword), &l)) != FT_OK) {
//			printf("iso9141_init: keyword recv err rc=%d\n", rc);
			break;
		}
		if (l != 2) {
//			printf("iso9141_init: keyword recv err l=%d\n", l);
			rc = FT_OTHER_ERROR;		// Time out
			break;
		}
		
		// Send key word 2
		usleep(25000);
		c = ~keyword[1];
		if ((rc = FT_Write(fd, &c, 1, &l)) != FT_OK) {
//			printf("iso9141_init: ~kw2 send err rc=%d\n", rc);
			break;
		}
		if ((rc = FT_Read(fd, &c, 1, &l)) != FT_OK) {	// Dummy recv
//			printf("iso9141_init: ~kw2 dummy recv err rc=%d\n", rc);
			break;
		}
		
		// Recieve my address
		if ((rc = FT_Read(fd, &c, 1, &l)) != FT_OK) {
//			printf("iso9141_init: ~addrrecv err rc=%d\n", rc);
			break;
		}
		if (c != (unsigned char)(~addr)) {
//			printf("iso9141_init: ERR: ~addr=%02x l=%d\n", c, l);
			rc = FT_OTHER_ERROR;		// Protocol error
			break;
		}
		
		break;
		
	} while(1);
	
	return((int)rc);
};

int iso9141_send_msg(unsigned char *data, size_t len){
	FT_STATUS	rc;
	unsigned char	head[3] = {0x68, 0x6a, 0xf1};
	unsigned char	csum[1];
	int	s;
	int	i;
	
	
	// Calc check sum
	s = 0;
	for(i = 0; i < sizeof(head); i++) {
		s += head[i];
	}
	for(i = 0; i < len; i++) {
		s += data[i];
	}
	csum[0] = s;
	
	// Debug
#ifdef DEBUG
	printf("iso9141_send: dump=");
	for(i = 0; i < sizeof(head); i++) {
		printf("%02x", head[i]);
	}
	printf(" ");
	for(i = 0; i < len; i++) {
		printf("%02x", data[i]);
	}
	printf(" %02x\n", csum[0]);
#endif
	
	// Send 3 part timming
	usleep(P3_MIN);
	do {
		DWORD	l;
		unsigned char	dummy[8];
		
		if ((rc = FT_Write(fd, head, sizeof(head), &l)) != FT_OK) {
//			printf("iso9141_send: haed send err rc=%d\n", rc);
			break;
		}
		if ((rc = FT_Read(fd, dummy, sizeof(head), &l)) != FT_OK) {
//			printf("iso9141_send: haed dmyrcv err rc=%d\n", rc);
			break;
		}
		usleep(P4_MIN);
		if ((rc = FT_Write(fd, data, len, &l)) != FT_OK) {
//			printf("iso9141_send: data send err rc=%d\n", rc);
			break;
		}
		if ((rc = FT_Read(fd, dummy, len, &l)) != FT_OK) {
//			printf("iso9141_send: data dmyrcv err rc=%d\n", rc);
			break;
		}
		usleep(P4_MIN);
		if ((rc = FT_Write(fd, csum, sizeof(csum), &l)) != FT_OK) {
//			printf("iso9141_send: csum send err rc=%d\n", rc);
			break;
		}
		if ((rc = FT_Read(fd, dummy, sizeof(csum), &l)) != FT_OK) {
//			printf("iso9141_send: csum dmyrcv err rc=%d\n", rc);
			break;
		}
		break;
	} while(1);
	
	return((int)rc);
}

int iso9141_recv_response(unsigned char *data, size_t *len){
	FT_STATUS	rc;
	
	*len = 0;
	do {
		DWORD	l;
		unsigned char	buf[11];
		int	s;
		int	i;
		
		// Recv
		if ((rc = FT_Read(fd, buf, sizeof(buf), &l)) != FT_OK) {
//			printf("iso9141_recv: read err rc=%d\n", rc);
			break;
		}
		if (l <= 1) {
			rc = FT_OTHER_ERROR;	// Time out
//			printf("iso9141_recv: read time out rc=%d len=%d\n", rc, l);
			break;
		}
		
		// Debug
#ifdef DEBUG
		printf("iso9141_recv: dump=");
		for(i = 0; i < (l-1); i++) {
			if (i == 3) {
				printf(" ");
			}
			printf("%02x", buf[i]);
		}
		printf(" %02x\n", buf[l-1]);
#endif
		
		// Sum check
		if (l > 1) {
			s = 0;
			--l;
			for(i = 0; i < l; i++) {
				s += buf[i];
			}
			if ((unsigned char)s != buf[l]) {
				rc = FT_OTHER_ERROR;	// sum error
				break;
			}
		}
		
		// Set response data
		for(i = 3; i < l; i++) {
			*data++ = buf[i];
			(*len)++;
		}
		break;
		
	} while(1);
	
	return((int)rc);
}

